#ifndef S3D_ARRAY_OBJECT_H
#define S3D_ARRAY_OBJECT_H

#include <s3d/array.h>

namespace Sengir3D{

class NoInitobject{};
template<typename inittype,Enum initmode> class Objectarrayhelper{};
template<typename inittype> class Objectarrayhelper<inittype,NONE>{
	public:    template<typename T> inline void ConstructObject(T &) const{}
	           template<typename T> inline void DestructObject(T &) const{}
	           inline inittype GetInitobject() const{ return NoInitobject(); }
	           inline Objectarrayhelper(const inittype &){}
};
template<typename inittype> class Objectarrayhelper<inittype,NORMAL>{
	private:   inittype initobject;
	public:    template<typename T> inline void ConstructObject(T &object) const{ new (&object) T(initobject); }
	           template<typename T> inline void DestructObject(T &object) const{ object.~T(); }
	           inline inittype GetInitobject() const{ return initobject; }
             inline Objectarrayhelper(const inittype &newinitobject) : initobject(newinitobject){}
};
template<> class Objectarrayhelper<NoInitobject,NORMAL>{
	public: template<typename T> inline void ConstructObject(T &object) const{ new (&object) T; }
	        template<typename T> inline void DestructObject(T &object) const{ object.~T(); }
	        inline NoInitobject GetInitobject() const{ return NoInitobject(); }
	        inline Objectarrayhelper(const NoInitobject &){}
};
template<typename inittype> class Objectarrayhelper<inittype,CALL>{
	private:   inittype initobject;
	public:    template<typename T> inline void ConstructObject(T &object) const{ object.Construct(initobject); }
	           template<typename T> inline void DestructObject(T &object) const{ object.Destruct(); }
	           inline inittype GetInitobject() const{ return initobject; }
	           inline Objectarrayhelper(const inittype &newinitobject) : initobject(newinitobject){}
};
template<> class Objectarrayhelper<NoInitobject,CALL>{
	public: template<typename T> inline void ConstructObject(T &object) const{ object.Construct(); }
	        template<typename T> inline void DestructObject(T &object) const{ object.Destruct(); }
	        inline NoInitobject GetInitobject() const{ return NoInitobject(); }
	        inline Objectarrayhelper(const NoInitobject &){}
};


template<class T,Enum initmode,typename inittype,Boolean staticposition> class Objectarraybase{};
template<class T,Enum initmode,typename inittype> class Objectarraybase<T,initmode,inittype,NO> : protected Array<T>,protected Objectarrayhelper<inittype,initmode>{
	protected: void Reallocate(uint newallocated){
	           	element=(T *)realloc(element,sizeof(T)*newallocated);
	           	assert(element);
	           	if(initmode!=NORMAL) for(uint c=allocated;c<newallocated;c++) new (element+c) T;
	           	allocated=newallocated;
	           }
	public:    inline void SetNumber(uint newnumber){
	           uint c;
	           	if(newnumber > number){
	           		if(newnumber > allocated) Reallocate(newnumber);
	           		for(c=number;c<newnumber;c++) ConstructObject(element[c]);
	           	}else{
	           		for(c=newnumber;c<number;c++) DestructObject(element[c]);
	           	}
	           	number=newnumber;
	           }
	protected: void Free(){
	           	if(initmode!=NORMAL) for(uint c=0;c<allocated;c++) element[c].~T();
	           	allocated=0;
	           	if(element){
	           		free(element);
	           		element=NULL;
	           	}
	           }
	public:    inline const T &operator[](uint i) const{
	           	assert(i < number);
	           	return element[i];
	           }
	           inline T &operator[](uint i){
	           	assert(i < number);
	           	return element[i];
	           }
	           inline void Exchange(Reference ref1,Reference ref2){ Array<T>::Exchange(ref1,ref2); }
	           inline void RemoveElement(uint i){
	           	Exchange(i,number-1);
	           	SetNumber(number-1);
	           }
	           void Remove(const T &oldelement){
	           	for(int c=number-1;c>=0;c--){
	           		if(element[c]==oldelement){
	           			RemoveElement(c);
	           			return;
	           		}
	           	}
	           	throw ElementNotFound();
	           }
	           inline Objectarraybase(const inittype &newinitobject=NoInitobject()) : Objectarrayhelper<inittype,initmode>(newinitobject){}
};
template<class T,Enum initmode,typename inittype> class Objectarraybase<T,initmode,inittype,YES> : protected Array<T *>,protected Objectarrayhelper<inittype,initmode>{
	protected: void Reallocate(uint newallocated){
	           uint c;
	           	element=(T **)realloc(element,sizeof(T *)*newallocated);
	           	assert(element);
	           	if(initmode!=NORMAL) for(c=allocated;c<newallocated;c++) element[c]=new T;
	           	else for(c=allocated;c<newallocated;c++) element[c]=(T *)malloc(sizeof(T));
	           	allocated=newallocated;
	           }
	public:    inline void SetNumber(uint newnumber){
	           uint c;
	           	if(newnumber > number){
	           		if(newnumber > allocated) Reallocate(newnumber);
	           		for(c=number;c<newnumber;c++) ConstructObject(*(element[c]));
	           	}else{
	           		for(c=newnumber;c<number;c++) DestructObject(*(element[c]));
	           	}
	           	number=newnumber;
	           }
	protected: void Free(){
	           uint c;
	           	if(initmode!=NORMAL) for(c=0;c<allocated;c++) delete element[c];
	           	else for(c=0;c<allocated;c++) free(element[c]);
	           	allocated=0;
	           	if(element){
	           		free(element);
	           		element=NULL;
	           	}
	           }
	public:    inline const T &operator[](uint i) const{
	           	assert(i < number);
	           	return *(element[i]);
	           }
	           inline T &operator[](uint i){
	           	assert(i < number);
	           	return *(element[i]);
	           }
	           inline Objectarraybase(const inittype &newinitobject=NoInitobject()) : Objectarrayhelper<inittype,initmode>(newinitobject){}
};

template<class T,Enum initmode,typename inittype,Boolean staticposition> class Objectarray : public Objectarraybase<T,initmode,inittype,staticposition>{
	public:    inline uint GetNumber() const{ return number; }
	           inline void SetMinNumber(uint minnumber){
	           	if(minnumber > number) SetNumber(minnumber);
	           }
	           inline void Clear(){
	           	for(uint c=0;c<number;c++) DestructObject(operator[](c));
	           	number=0;
	           }
	           void Free(){
	           	Clear();
	           	Objectarraybase<T,initmode,inittype,staticposition>::Free();
	           }
	           inline void Construct(){ number=0; }
	           inline void Destruct(){ Clear(); }
	           inline void Add(const T &newelement){
	           	SetNumber(number+1);
	           	operator[](number-1)=newelement;
	           }
	           template<Enum initmode2,typename inittype2,Boolean staticposition2> void Add(const Objectarray<T,initmode2,inittype2,staticposition2> &addarray){
	           	for(uint c=0;c<addarray.GetNumber();c++) Add(addarray[c]);
	           }
	           inline void operator++(int){ SetNumber(number+1); }
	           inline void operator--(int){
	           	assert(number>0);
	           	SetNumber(number-1);
	           }
	           inline void operator+=(uint add){ SetNumber(number+add); }
	           inline void operator-=(uint sub){
	           	assert(number >= sub);
	           	SetNumber(number-sub);
	           }
	           inline Reference GetReference(const T &element,Enum parameter=REQUIRED) const{
	           	for(int c=number-1;c>=0;c--) if(operator[](c)==element) return c;
	           	if(parameter==REQUIRED) throw ElementNotFound(); else return INVALID;
	           }
	           inline Boolean Contains(const T &element) const{
	           	for(int c=number-1;c>=0;c--) if(operator[](c)==element) return YES;
	           	return NO;
	           }
	           Objectarray<T,initmode,inittype,staticposition> &operator=(const Objectarray<T,initmode,inittype,staticposition> &newarray){
	           	if(&newarray!=this){
	           		SetNumber(newarray.GetNumber());
	           		for(uint c=0;c<number;c++) operator[](c)=newarray[c];
	           	}
	           	return *this;
	           }
	           template<Enum initmode2,typename inittype2,Boolean staticposition2> Objectarray<T,initmode,inittype,staticposition> &operator=(const Objectarray<T,initmode2,inittype2,staticposition2> &newarray){
	           	if(&newarray!=this){
	           		SetNumber(newarray.GetNumber());
	           		for(uint c=0;c<number;c++) operator[](c)=newarray[c];
	           	}
	           }
	           explicit inline Objectarray(const inittype &newinitobject=NoInitobject()) : Objectarraybase<T,initmode,inittype,staticposition>(newinitobject) {}
	           Objectarray(const Objectarray<T,initmode,inittype,staticposition> &newarray) : Objectarraybase<T,initmode,inittype,staticposition>(newarray.GetInitobject()){ operator=(newarray); }
	           template<Enum initmode2,typename inittype2,Boolean staticposition2> Objectarray(const Objectarray<T,initmode2,inittype2,staticposition2> &newarray) : Objectarraybase<T,initmode,inittype,staticposition>(newarray.GetInitobject()){ operator=(newarray); }
	           template<Enum initmode2,typename inittype2,Boolean staticposition2> Objectarray(const Objectarray<T,initmode2,inittype2,staticposition2> &newarray,const inittype &newinitobject) : Objectarraybase<T,initmode,inittype,staticposition>(newinitobject){ operator=(newarray); }
	           inline ~Objectarray(){ Free(); }
};

}
#endif

